package checkers.nullness;

import java.util.Collections;
import java.util.List;

import javax.lang.model.element.*;

import javax.lang.model.type.*;

import checkers.basetype.BaseTypeChecker;
import checkers.flow.*;
import checkers.nullness.quals.*;
import checkers.quals.*;
import checkers.types.*;
import checkers.types.AnnotatedTypeMirror.*;
import checkers.types.visitors.AnnotatedTypeScanner;
import checkers.util.*;

import com.sun.source.tree.*;
import com.sun.source.util.*;

/**
 * Adds the {@link NonNull} annotation to a type that is:
 *
 * 1. in the scope of a {@link DefaultQualifier} annotation and matches its location
 *    criteria
 * 2. determined to be {@link NonNull} by flow-sensitive inference
 * 3. the class in a static member access (e.g., "System" in "System.out")
 * 4. an array-creation expression (with new)
 * 5. an object-creation expression (with new)
 * 6. a string literal
 * 7. a package declaration
 * 8. an exception parameter
 * 9. the receiver type of a non-static (and non-constructor) method
 *
 * Adds the {@link Raw} annotation to
 * 10. the receiver type of constructors.
 *
 * <p>
 *
 * Additionally, the type factory will add the {@link Nullable} annotation to a
 * type if the input is
 * 11. the null literal
 * 12. of type {@link Void},
 * 13. may resolve the types of some {@link NonNull} fields as {@link Nullable}
 *     depending on the presence of a {@link Raw} annotation on a constructor
 *     or method receiver.  Please review
 *     <a href="http://people.csail.mit.edu/mernst/pubs/pluggable-checkers-issta2008-abstract.html">
 *     the Checker Framework</a> for {@link Raw} semantics.
 */
public class NullnessAnnotatedTypeFactory extends AnnotatedTypeFactory {

    private final Flow flow;
    private final QualifierDefaults defaults;
    private final TypeAnnotator typeAnnotator;
    private final TreeAnnotator treeAnnotator;
    private final QualifierPolymorphism poly;

    private final AnnotationCompleter completer = new AnnotationCompleter();

    /** Represents the Nullness Checker qualifiers */
    protected final AnnotationMirror NONNULL, RAW, NULLABLE;

    /** Creates a {@link NullnessAnnotatedTypeFactory}. */
    public NullnessAnnotatedTypeFactory(NullnessChecker checker,
            CompilationUnitTree root) {
        super(checker, root);

        typeAnnotator = new NonNullTypeAnnotator(checker);
        treeAnnotator = new NonNullTreeAnnotator(checker);

        NONNULL = this.annotations.fromClass(NonNull.class);
        RAW = this.annotations.fromClass(Raw.class);
        NULLABLE = this.annotations.fromClass(Nullable.class);

        defaults = new QualifierDefaults(this, this.annotations);
        defaults.setAbsoluteDefaults(NONNULL, Collections.singleton(DefaultLocation.ALL_EXCEPT_LOCALS));

        this.poly = new QualifierPolymorphism(checker, this);

        flow = new NullnessFlow(checker, root, this);
        flow.scan(root, null);
    }

    @Override
    protected void annotateImplicit(Element elt, AnnotatedTypeMirror type) {

        if (elt instanceof VariableElement)
            annotateIfStatic(elt, type);
        typeAnnotator.visit(type);
        defaults.annotate(elt, type);
        completer.visit(type);
    }

    @Override
    protected void annotateImplicit(Tree tree, AnnotatedTypeMirror type) {
        treeAnnotator.visit(tree, type);
        typeAnnotator.visit(type);
        defaults.annotate(tree, type);
        substituteRaw(tree, type);
        // case 2. inference
        final AnnotationMirror inferred = flow.test(tree);
        if (inferred != null) {
            type.clearAnnotations();
            type.addAnnotation(inferred);
        }
        completer.visit(type);
    }

    @Override
    protected AnnotatedDeclaredType getImplicitReceiverType(Tree tree) {
        AnnotatedDeclaredType type = super.getImplicitReceiverType(tree);
        // 'this' should always be nonnull, unless it's raw
        if (type != null && !type.hasAnnotation(RAW)) {
            type.clearAnnotations();
            type.addAnnotation(NONNULL);
        }
        return type;
    }

    @Override
    protected void postDirectSuperTypes(AnnotatedTypeMirror type,
            List<? extends AnnotatedTypeMirror> supertypes) {
        for (AnnotatedTypeMirror supertype : supertypes) {
            typeAnnotator.visit(supertype);
            if (supertype.getKind() == TypeKind.DECLARED)
                defaults.annotateTypeElement((TypeElement)((AnnotatedDeclaredType)supertype).getUnderlyingType().asElement(), supertype);
            completer.visit(supertype);
        }
    }

    @Override
    public AnnotatedExecutableType methodFromUse(MethodInvocationTree tree) {
        AnnotatedExecutableType method = super.methodFromUse(tree);
        poly.annotate(tree, method);
        poly.annotate(method.getElement(), method);
        return method;
    }

    /**
     * Substitutes {@link Raw} annotations on a type. If the the receiver of
     * the member select expression tree (which might implicitly be "this") is
     * {@link Raw}, this replaces the annotations on {@code type} with
     * {@link Raw}.
     *
     * @param tree a select expression, which might be implicit
     * @param type the type for which to substitute annotations based on the
     *        placement of {@link Raw}
     * @return true iff we reduced the annotated type from {@link NonNull} to
     *         {@link Raw}
     */
    boolean substituteRaw(Tree tree, AnnotatedTypeMirror type) {

        // If it's not an expression tree, it's definitely not a select.
        if (!(tree instanceof ExpressionTree))
            return false;

        // Identifiers might implicitly select from "this", but not if they're
        // class or interface identifiers.
        if (tree instanceof IdentifierTree) {

            Element elt = TreeUtils.elementFromUse((IdentifierTree) tree);
            if (elt == null || !elt.getKind().isField())
                return false;

            Tree decl = declarationFromElement(elt);
            if (this.getPath(decl) != null
                        && decl instanceof VariableTree
                        && ((VariableTree) decl).getInitializer() != null
                        && getAnnotatedType(((VariableTree) decl).getInitializer()).hasAnnotation(NONNULL))
            return false;
        }

        final AnnotatedTypeMirror select = getReceiver((ExpressionTree) tree);
        if (select != null && select.hasAnnotation(RAW)
                && !type.hasAnnotation(NULLABLE) && !type.getKind().isPrimitive()) {
            boolean wasNN = type.hasAnnotation(NONNULL);
            type.clearAnnotations();
            type.addAnnotation(RAW);
            return wasNN;
        }
        return false;
    }

    /**
     * If the element is {@link NonNull} when used in a static member access,
     *  modifies the element's type (by adding {@link NonNull}).
     *
     * @param elt the element being accessed
     * @param type the type of the element {@code elt}
     */
    private void annotateIfStatic(Element elt, AnnotatedTypeMirror type) {

        if (elt == null)
            return;

        if (elt.getKind().isClass() || elt.getKind().isInterface()
                // Workaround for System.{out,in,err} issue: assume all static
                // fields in java.lang.System are nonnull.
                || isSystemField(elt)) {
            type.clearAnnotations();
            type.addAnnotation(NONNULL);
        }
    }

    private boolean isSystemField(Element elt) {
        if (!elt.getKind().isField())
            return false;

        if (!ElementUtils.isStatic(elt) || !ElementUtils.isFinal(elt))
            return false;

        VariableElement var = (VariableElement)elt;

        // Heuristic: if we have a static final field in a system package,
        // treat it as NonNull (many like Boolean.TYPE and System.out
        // have constant value null but are set by the VM).
        boolean inJavaPackage = ElementUtils.getQualifiedClassName(var).toString().startsWith("java.");

        return (var.getConstantValue() != null
                || var.getSimpleName().contentEquals("class")
                || inJavaPackage);
    }

    /**
     * Ensures that all types are fully annotated, and no type are unqualified,
     * and that no type can have multiple qualifiers
     */
    private class AnnotationCompleter extends AnnotatedTypeScanner<Void, Void> {

        @Override
        protected Void scan(AnnotatedTypeMirror type, Void p) {

            if (type == null)
                return super.scan(type, p);

            if (type.getAnnotations().isEmpty())
                type.addAnnotation(NULLABLE);
            else if (type.hasAnnotation(RAW))
                type.removeAnnotation(NONNULL);
            else if (type.hasAnnotation(NONNULL))
                type.removeAnnotation(NULLABLE);

            if (type instanceof AnnotatedDeclaredType) {
                Name name = TypesUtils.getQualifiedName((DeclaredType)type.getUnderlyingType());
                if (name.contentEquals("java.lang.Void")) {
                    type.clearAnnotations();
                    type.addAnnotation(NULLABLE);
                }
            }

            assert !type.getAnnotations().isEmpty() : type;

            return super.scan(type, p);
        }
    }

    /**
     * Adds {@link NonNull} and {@link Raw} annotations to the type based on
     * the underlying (unannotated) type.
     */
    private class NonNullTypeAnnotator extends TypeAnnotator {

        /** Creates a {@link NonNullTypeAnnotator} for the given checker. */
        NonNullTypeAnnotator(BaseTypeChecker checker) {
            super(checker);
        }

        /**
         * cases 7, 10.
         * Adds {@link NonNull} and {@link Raw} annotations to the unannotated
         * method receiver
         */
        @Override
        public Void visitExecutable(AnnotatedExecutableType type, ElementKind p) {

            // Don't add implicit receiver annotations if some are already there.
            if (!type.getReceiverType().getAnnotations().isEmpty())
                return super.visitExecutable(type, p);

            ExecutableElement elt = type.getElement();
            assert elt != null;

            // Add @Raw to constructors.
            if (elt.getKind() == ElementKind.CONSTRUCTOR)
                type.getReceiverType().addAnnotation(RAW);

            // Add @NonNull to non-static non-constructors.
            if (!ElementUtils.isStatic(elt))
                type.getReceiverType().addAnnotation(NONNULL);

            return super.visitExecutable(type, p);
        }
    }

    /**
     * Adds {@link NonNull} annotations to a type based on the AST from which
     * the type was obtained.
     */
    private class NonNullTreeAnnotator extends TreeAnnotator {

        /** Creates a {@link NonNullTreeAnnotator} for the given checker. */
        NonNullTreeAnnotator(BaseTypeChecker checker) {
            super(checker);
        }

        @Override
        public Void visitMemberSelect(MemberSelectTree node, AnnotatedTypeMirror type) {

            Element elt = TreeUtils.elementFromUse(node);
            assert elt != null;
            // case 3. static method access
            annotateIfStatic(elt, type);
            return super.visitMemberSelect(node, type);
        }

        @Override
        public Void visitIdentifier(IdentifierTree node, AnnotatedTypeMirror type) {

            Element elt = TreeUtils.elementFromUse(node);
            assert elt != null;

            // Workaround: exception parameters should be implicitly
            // NonNull, but due to a compiler bug they have
            // kind PARAMETER instead of EXCEPTION_PARAMETER. Until it's
            // fixed we manually inspect enclosing catch blocks.
            // case 8. exception parameter
            if (isExceptionParameter(node))
                type.addAnnotation(NONNULL);

            // case 3. static method access
            annotateIfStatic(elt, type);
            return super.visitIdentifier(node, type);
        }

        private boolean isExceptionParameter(IdentifierTree node) {
            Element elt = TreeUtils.elementFromUse(node);
            assert elt != null;
            if (elt.getKind() != ElementKind.PARAMETER
                    || !TypesUtils.isThrowable(elt.asType()))
                return false;
            final TreePath path = getPath(node);
            CatchTree ct = (CatchTree)TreeUtils.enclosingOfKind(path, Tree.Kind.CATCH);
            if (ct == null)
                return false;

            final VariableTree catchParamTree = ct.getParameter();
            final VariableElement catchParamElt = TreeUtils.elementFromDeclaration(catchParamTree);
            return elt.equals(catchParamElt);
        }
    }
}
