package checkers.nullness;

import javax.lang.model.element.*;
import javax.lang.model.type.TypeKind;

import checkers.basetype.*;
import checkers.nullness.quals.NonNull;
import checkers.source.Result;
import checkers.types.AnnotatedTypeMirror;
import checkers.types.AnnotatedTypeMirror.*;
import checkers.util.TreeUtils;
import static checkers.util.TypesUtils.isBoxedPrimitive;

import com.sun.source.tree.*;

/**
 * A type-checking visitor for the Nullness type system.
 * This visitor reports errors ("dereference.of.nullable") or
 * warnings for violations for the following cases:
 *
 * 1. if the receiver of a member dereference is not NonNull
 * 2. if the receiver of an implicit ".iterator()" dereference in an enhanced
 *    for loop is not NonNull
 * 3. if an accessed array is not NonNull
 *
 * Self-accesses (via {@code this} or {@code super}) can always be dereferenced.
 *
 * @see NullnessChecker
 */
public class NullnessVisitor extends BaseTypeVisitor<Void, Void> {

    /** The {@link NonNull} annotation */
    private final AnnotationMirror NONNULL;

    /**
     * Creates a new visitor for type-checking {@link NonNull}.
     *
     * @param checker the checker to use
     * @param root the root of the input program's AST to check
     */
    public NullnessVisitor(NullnessChecker checker, CompilationUnitTree root) {
        super(checker, root);
        NONNULL = this.annoFactory.fromClass(NonNull.class);
    }

    @Override
    public Void visitMemberSelect(MemberSelectTree node, Void p) {
        // Case 1. dereferencing a nullable value
        if (!TreeUtils.isSelfAccess(node))
            checkForNullability(node.getExpression());

        return super.visitMemberSelect(node, p);
    }

    @Override
    public Void visitEnhancedForLoop(EnhancedForLoopTree node, Void p) {
        // case 2. implicit dereferencing in enhanced for loop trees
        checkForNullability(node.getExpression());
        return super.visitEnhancedForLoop(node, p);
    }

    @Override
    public Void visitArrayAccess(ArrayAccessTree node, Void p) {
        // case 3. implicit dereferencing in array access
        checkForNullability(node.getExpression());
        return super.visitArrayAccess(node, p);
    }

    @Override
    public Void visitThrow(ThrowTree node, Void p) {
        checkForNullability(node.getExpression());
        return super.visitThrow(node, p);
    }

    @Override
    public Void visitSynchronized(SynchronizedTree node, Void p) {
        checkForNullability(node.getExpression());
        return super.visitSynchronized(node, p);
    }

    @Override
    public Void visitUnary(UnaryTree node, Void p) {
        checkForNullability(node.getExpression());
        return super.visitUnary(node, p);
    }

    @Override
    public Void visitBinary(BinaryTree node, Void p) {
        if (node.getKind() != Tree.Kind.EQUAL_TO
                && node.getKind() != Tree.Kind.NOT_EQUAL_TO
                && isPracticallyPrimitive(node)) {
            checkForNullability(node.getLeftOperand());
            checkForNullability(node.getRightOperand());
        }
        return super.visitBinary(node, p);
    }

    @Override
    public Void visitCompoundAssignment(CompoundAssignmentTree node, Void p) {
        if (isPracticallyPrimitive(node)) {
            checkForNullability(node.getVariable());
            checkForNullability(node.getExpression());
        }
        return super.visitCompoundAssignment(node, p);
    }

    private final boolean isPracticallyPrimitive(ExpressionTree tree) {
        AnnotatedTypeMirror type = atypeFactory.fromExpression(tree);
        return (type.getKind().isPrimitive()
                || isBoxedPrimitive(type.getUnderlyingType()));
    }

    @Override
    protected boolean checkMethodInvocability(AnnotatedExecutableType method,
            MethodInvocationTree node) {
        // Claim that methods with a @NonNull receiver are invokable so that
        // visitMemberSelect issues dereference errors instead.
        if (!TreeUtils.isSelfAccess(node)
                && method.getReceiverType().hasAnnotation(NONNULL))
            return true;
        return super.checkMethodInvocability(method, node);
    }

    /**
     * Issues a 'dereference.of.nullable' if the type is not of a
     * {@link NonNull} type.
     *
     * @param type  type to be checked nullability
     * @param tree  the tree where the error is to reported
     */
    private void checkForNullability(ExpressionTree tree) {
        AnnotatedTypeMirror type = atypeFactory.getAnnotatedType(tree);
        if (!type.hasAnnotation(NONNULL))
            checker.report(Result.failure("dereference.of.nullable", tree), tree);
    }
}
