package checkers.util.skel;

import java.io.InputStream;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.annotation.processing.ProcessingEnvironment;
import javax.lang.model.element.*;
import javax.lang.model.util.ElementFilter;
import javax.lang.model.util.Elements;

import checkers.types.AnnotatedTypeFactory;
import checkers.types.AnnotatedTypeMirror;
import checkers.types.AnnotatedTypeMirror.*;
import checkers.util.AnnotationUtils;

import japa.parser.JavaParser;
import japa.parser.ast.*;
import japa.parser.ast.body.*;
import japa.parser.ast.expr.AnnotationExpr;
import japa.parser.ast.type.*;

public class IndexParser {
    final IndexUnit index;
    final AnnotatedTypeFactory atypeFactory;
    final AnnotationUtils annoUtils;
    final ProcessingEnvironment env;
    final Elements elements;

    final Map<String, AnnotationMirror> annotations;

    public IndexParser(InputStream inputStream, AnnotatedTypeFactory factory, ProcessingEnvironment env) {
        try {
            this.index = JavaParser.parse(inputStream);
        } catch (Exception e) {
            throw new Error(e);
        }
        this.atypeFactory = factory;
        this.env = env;
        this.annoUtils = new AnnotationUtils(env);
        this.elements = env.getElementUtils();
        annotations = getSupportedAnnotations();
    }

    private Map<String, AnnotationMirror> getSupportedAnnotations() {
        assert !index.getCompilationUnits().isEmpty();
        CompilationUnit cu = index.getCompilationUnits().get(0);
        AnnotationUtils annoUtils = new AnnotationUtils(env);

        Map<String, AnnotationMirror> result = new HashMap<String, AnnotationMirror>();

        for (ImportDeclaration importDecl : cu.getImports()) {
            String imported = importDecl.getName().toString();
            try {
                AnnotationMirror anno = annoUtils.fromName(imported);
                result.put(anno.getAnnotationType().asElement().getSimpleName().toString(), anno);
            } catch (AssertionError error) {
                // do nothing
            }
        }
        return result;
    }

    public Map<Element, AnnotatedTypeMirror> parse() {
        Map<Element, AnnotatedTypeMirror> result = new HashMap<Element, AnnotatedTypeMirror>();
        parse(this.index, result);
        return result;
    }

    private void parse(IndexUnit index, Map<Element, AnnotatedTypeMirror> result) {
        for (CompilationUnit cu : index.getCompilationUnits())
            parse(cu, result);
    }

    private void parse(CompilationUnit cu, Map<Element, AnnotatedTypeMirror> result) {
        final String packageName;
        if (cu.getPakage() == null)
            packageName = null;
        else
            packageName = cu.getPakage().getName().toString();
        if (cu.getTypes() != null) {
            for (TypeDeclaration typeDecl : cu.getTypes())
                parse(typeDecl, packageName, result);
        }
    }

    private void parse(TypeDeclaration type, String packageName, Map<Element, AnnotatedTypeMirror> result) {
        String typeName = (packageName == null ? "" : packageName + ".") + type.getName();
        TypeElement typeElt = elements.getTypeElement(typeName);
        // couldn't find type.  not in class path
        // TODO: Should throw exception?!
        if (typeElt == null
                || typeElt.getKind() == ElementKind.ENUM
                || typeElt.getKind() == ElementKind.ANNOTATION_TYPE)
            return;

        if (type instanceof ClassOrInterfaceDeclaration)
            parseType((ClassOrInterfaceDeclaration)type, typeElt, result);

        Map<Element, BodyDeclaration> elementsToDecl = mapMembers(typeElt, type);
        for (Map.Entry<Element, BodyDeclaration> entry : elementsToDecl.entrySet()) {
            final Element elt = entry.getKey();
            final BodyDeclaration decl = entry.getValue();
            if (elt.getKind().isField())
                parseField((FieldDeclaration)decl, (VariableElement)elt, result);
            else if (elt.getKind() == ElementKind.CONSTRUCTOR)
                parseConstructor((ConstructorDeclaration)decl, (ExecutableElement)elt, result);
            else if (elt.getKind() == ElementKind.METHOD)
                parseMethod((MethodDeclaration)decl, (ExecutableElement)elt, result);
            else { /* do nothing */ }
        }
    }

    private void parseType(ClassOrInterfaceDeclaration decl, TypeElement elt, Map<Element, AnnotatedTypeMirror> result) {
        AnnotatedDeclaredType type = atypeFactory.fromElement(elt);
        annotate(type, decl.getAnnotations());
        result.put(elt, type);
    }

    private void parseMethod(MethodDeclaration decl, ExecutableElement elt,
            Map<Element, AnnotatedTypeMirror> result) {
        AnnotatedExecutableType methodType = atypeFactory.fromElement(elt);
        annotate(methodType.getReturnType(), decl.getType());

        for (int i = 0; i < methodType.getParameterTypes().size(); ++i) {
            AnnotatedTypeMirror paramType = methodType.getParameterTypes().get(i);
            Parameter param = decl.getParameters().get(i);
            annotate(paramType, param.getType());
        }

        annotate(methodType.getReceiverType(), decl.getReceiverAnnotations());

        result.put(elt, methodType);
    }

    private void annotate(AnnotatedTypeMirror atype, Type typeDef) {
        if (typeDef.getAnnotations() != null)
            annotate(atype, typeDef.getAnnotations());
        // TODO: Type bounds, type arguments, etc
    }

    private void parseConstructor(ConstructorDeclaration decl,
            ExecutableElement elt, Map<Element, AnnotatedTypeMirror> result) {
        AnnotatedExecutableType methodType = atypeFactory.fromElement(elt);

        for (int i = 0; i < methodType.getParameterTypes().size(); ++i) {
            AnnotatedTypeMirror paramType = methodType.getParameterTypes().get(i);
            Parameter param = decl.getParameters().get(i);
            annotate(paramType, param.getType());
        }

        annotate(methodType.getReceiverType(), decl.getReceiverAnnotations());

        result.put(elt, methodType);
    }

    private void parseField(FieldDeclaration decl,
            VariableElement elt, Map<Element, AnnotatedTypeMirror> result) {
        AnnotatedTypeMirror fieldType = atypeFactory.fromElement(elt);
        annotate(fieldType, decl.getType());
        result.put(elt, fieldType);
    }

    private void annotate(AnnotatedTypeMirror type, List<AnnotationExpr> annotations) {
        if (annotations == null)
            return;
        for (AnnotationExpr annotation : annotations) {
            String annoName = IndexUtil.getAnnotationName(annotation);
            AnnotationMirror annoMirror = this.annotations.get(annoName);
            if (annoMirror != null)
                type.addAnnotation(annoMirror);
        }
    }

    public Map<Element, BodyDeclaration> mapMembers(TypeElement typeElt, TypeDeclaration typeDecl) {
        assert typeElt.getSimpleName().contentEquals(typeDecl.getName());

        Map<Element, BodyDeclaration> result = new HashMap<Element, BodyDeclaration>();

        for (BodyDeclaration member : typeDecl.getMembers()) {
            if (member instanceof MethodDeclaration) {
                Element elt = findElement(typeElt, (MethodDeclaration)member);
                result.put(elt, member);
            } else if (member instanceof ConstructorDeclaration) {
                Element elt = findElement(typeElt, (ConstructorDeclaration)member);
                result.put(elt, member);
            } else if (member instanceof FieldDeclaration) {
                FieldDeclaration fieldDecl = (FieldDeclaration)member;
                for (VariableDeclarator var : fieldDecl.getVariables())
                    result.put(findElement(typeElt, var), fieldDecl);
            }
        }
        result.remove(null);
        return result;
    }

    public ExecutableElement findElement(TypeElement typeElt, MethodDeclaration methodDecl) {
        final String wantedMethodName = methodDecl.getName();
        final int wantedMethodParams =
            (methodDecl.getParameters() == null) ? 0 :
                methodDecl.getParameters().size();
        final String wantedMethodString = IndexUtil.toString(methodDecl);
        for (ExecutableElement method : ElementFilter.methodsIn(typeElt.getEnclosedElements())) {
            // do hearustics first
            if (wantedMethodParams == method.getParameters().size()
                    && wantedMethodName.contentEquals(method.getSimpleName())
                    && IndexUtil.toString(method).equals(wantedMethodString))
                return method;
        }
        return null;
    }

    public ExecutableElement findElement(TypeElement typeElt, ConstructorDeclaration methodDecl) {
        final int wantedMethodParams =
            (methodDecl.getParameters() == null) ? 0 :
                methodDecl.getParameters().size();
        final String wantedMethodString = IndexUtil.toString(methodDecl);
        for (ExecutableElement method : ElementFilter.constructorsIn(typeElt.getEnclosedElements())) {
            // do hearustics first
            if (wantedMethodParams == method.getParameters().size()
                    && IndexUtil.toString(method).equals(wantedMethodString))
                return method;
        }
        return null;
    }

    public VariableElement findElement(TypeElement typeElt, VariableDeclarator variable) {
        final String fieldName = variable.getId().getName();
        for (VariableElement field : ElementFilter.fieldsIn(typeElt.getEnclosedElements())) {
            if (fieldName.contains(field.getSimpleName()))
                return field;
        }
        return null;
    }
}
